use crate::api;
use crate::core;
use crate::repositories;
use diesel::{
  r2d2::{ConnectionManager, Pool},
  PgConnection,
};

///
pub struct Synchronizer {
  pool: Pool<ConnectionManager<PgConnection>>,
  api_config: core::config::ApiConfig,
  synchronization_history_repository: repositories::synchronization_history::repository::Repository,
  balance_repository: repositories::balance::repository::Repository,
  member_repository: repositories::member::repository::Repository,
  member_api: api::member::api::Api,
  member_account_api: api::member_account::api::Api,
}

impl std::fmt::Debug for Synchronizer {
  fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
    f.debug_struct("Synchronizer of api.kgon.identity").finish()
  }
}

impl std::clone::Clone for Synchronizer {
  fn clone(&self) -> Self {
    Self {
      pool: self.pool.clone(),
      api_config: self.api_config.clone(),
      synchronization_history_repository:
        repositories::synchronization_history::repository::Repository::new(),
      balance_repository: repositories::balance::repository::Repository::new(),
      member_repository: repositories::member::repository::Repository::new(),
      member_api: api::member::api::Api::new(self.api_config.clone()),
      member_account_api: api::member_account::api::Api::new(self.api_config.clone()),
    }
  }
}

impl Synchronizer {
  ///
  pub fn new(
    pool: Pool<ConnectionManager<PgConnection>>,
    api_config: core::config::ApiConfig,
  ) -> Synchronizer {
    Synchronizer {
      pool,
      api_config: api_config.clone(),
      synchronization_history_repository:
        repositories::synchronization_history::repository::Repository::new(),
      balance_repository: repositories::balance::repository::Repository::new(),
      member_repository: repositories::member::repository::Repository::new(),
      member_api: api::member::api::Api::new(api_config.clone()),
      member_account_api: api::member_account::api::Api::new(api_config),
    }
  }

  pub async fn balance_for_user_all(&self) -> Result<(), Box<dyn std::error::Error>> {
    let start_at = (chrono::Utc::now()).timestamp();

    if let Err(e) = async {
      let conn = self.pool.get().expect("conn");

      let req = api::member::models::ListMembersRequest { group_key: None };
      let res = self.member_api.list_members(req).await?;

      let mut messages: Vec<String> = vec![];

      for u in res.users {
        let req = api::member_account::models::GetBalanceForUserRequest {
          username: u.site_username,
        };
        let res = match self.member_account_api.get_balance_for_user(req).await {
          Ok(r) => r,
          Err(e) => {
            messages.push(format!(
              "api.get_balance_for_user code: {}, msg: {:?}",
              e.code, e.msg
            ));
            continue;
          }
        };

        let modify_member = repositories::member::models::ModifyMemberForBalance {
          balance: res.balance,
          balance_bota: res.balance_bota,
          balance_sum: res.balance_sum,
          companies: res.companies,
        };

        if let Err(e) = self
          .member_repository
          .update_balance(&conn, u.id, &modify_member)
        {
          messages.push(format!("repository.update_balance error: {}", e));
        };
      }

      if !messages.is_empty() {
        return Err(api::core::models::Error {
          code: -1,
          msg: Some(serde_json::to_string(&messages).expect("json")),
        });
      }

      self
        .synchronization_history_repository
        .insert(
          &conn,
          &repositories::synchronization_history::models::NewSynchronizationHistory {
            item: repositories::synchronization::models::ITEM_BALANCE_USER.to_string(),
            start_at,
            complete_at: (chrono::Utc::now()).timestamp(),
            code: 0,
            message: None,
          },
        )
        .expect("synchronization_history insert");

      Ok::<(), api::core::models::Error>(())
    }
    .await
    {
      let conn = self.pool.get().expect("conn");

      self
        .synchronization_history_repository
        .insert(
          &conn,
          &repositories::synchronization_history::models::NewSynchronizationHistory {
            item: repositories::synchronization::models::ITEM_BALANCE_USER.to_string(),
            start_at,
            complete_at: (chrono::Utc::now()).timestamp(),
            code: e.code,
            message: e.msg,
          },
        )
        .expect("synchronization_history insert");
    }

    Ok(())
  }

  pub async fn balance_for_user(
    &self,
    member_id: uuid::Uuid,
    usename: &str,
  ) -> Result<(), Box<dyn std::error::Error>> {
    let start_at = (chrono::Utc::now()).timestamp();

    if let Err(e) = async {
      let conn = self.pool.get().expect("conn");

      let m = match self
        .member_repository
        .select_by_member_id(&conn, member_id)
        .expect("member_repository.select_by_member_id")
      {
        Some(m) => m,
        None => {
          return Ok(());
        }
      };

      let req = api::member_account::models::GetBalanceForUserRequest {
        username: usename.to_string(),
      };
      let res = self.member_account_api.get_balance_for_user(req).await?;

      let modify_member = repositories::member::models::ModifyMemberForBalance {
        balance: res.balance,
        balance_bota: res.balance_bota,
        balance_sum: res.balance_sum,
        companies: res.companies,
      };

      self
        .member_repository
        .update_balance(&conn, m.id, &modify_member)
        .expect("member_repository.update_balance");

      Ok::<(), api::core::models::Error>(())
    }
    .await
    {
      let conn = self.pool.get().expect("conn");

      self
        .synchronization_history_repository
        .insert(
          &conn,
          &repositories::synchronization_history::models::NewSynchronizationHistory {
            item: repositories::synchronization::models::ITEM_BALANCE_USER.to_string(),
            start_at,
            complete_at: (chrono::Utc::now()).timestamp(),
            code: e.code,
            message: e.msg,
          },
        )
        .expect("synchronization_history insert");
    }

    Ok(())
  }

  pub async fn balance_for_partner(&self) -> Result<(), Box<dyn std::error::Error>> {
    let start_at = (chrono::Utc::now()).timestamp();

    if let Err(e) = async {
      let conn = self.pool.get().expect("conn");

      let req = api::member_account::models::GetBalanceForPartnerRequest {};
      let res = self.member_account_api.get_balance_for_partner(req).await?;

      match self
        .balance_repository
        .select(&conn)
        .expect("balance select")
      {
        Some(b) => {
          self
            .balance_repository
            .update(
              &conn,
              b.id,
              &repositories::balance::models::ModifyBalance {
                balance: res.balance,
                balance_bota: res.balance_bota,
              },
            )
            .expect("balance update");
        }
        None => {
          self
            .balance_repository
            .insert(
              &conn,
              &repositories::balance::models::NewBalance {
                balance: res.balance,
                balance_bota: res.balance_bota,
              },
            )
            .expect("balance insert");
        }
      }

      self
        .synchronization_history_repository
        .insert(
          &conn,
          &repositories::synchronization_history::models::NewSynchronizationHistory {
            item: repositories::synchronization::models::ITEM_BALANCE_PARTNER.to_string(),
            start_at,
            complete_at: (chrono::Utc::now()).timestamp(),
            code: 0,
            message: None,
          },
        )
        .expect("synchronization_history insert");

      Ok::<(), api::core::models::Error>(())
    }
    .await
    {
      let conn = self.pool.get().expect("conn");
      self
        .synchronization_history_repository
        .insert(
          &conn,
          &repositories::synchronization_history::models::NewSynchronizationHistory {
            item: repositories::synchronization::models::ITEM_BALANCE_PARTNER.to_string(),
            start_at,
            complete_at: (chrono::Utc::now()).timestamp(),
            code: e.code,
            message: e.msg,
          },
        )
        .expect("synchronization_history insert");
    }

    Ok(())
  }
}
